package org.eclipse.jdt.internal.compiler.apt.model;

package .tmp;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.type.ErrorType;
import javax.lang.model.type.NoType;
import javax.lang.model.type.NullType;
import javax.lang.model.type.TypeKind;
import javax.lang.model.type.TypeMirror;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl;
import org.eclipse.jdt.internal.compiler.apt.model.ArrayTypeImpl;
import org.eclipse.jdt.internal.compiler.apt.model.DeclaredTypeImpl;
import org.eclipse.jdt.internal.compiler.apt.model.ErrorTypeElement;
import org.eclipse.jdt.internal.compiler.apt.model.ErrorTypeImpl;
import org.eclipse.jdt.internal.compiler.apt.model.ExecutableElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.ExecutableTypeImpl;
import org.eclipse.jdt.internal.compiler.apt.model.NoTypeImpl;
import org.eclipse.jdt.internal.compiler.apt.model.PackageElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.PrimitiveTypeImpl;
import org.eclipse.jdt.internal.compiler.apt.model.TypeElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.TypeParameterElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.TypeVariableImpl;
import org.eclipse.jdt.internal.compiler.apt.model.VariableElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.WildcardTypeImpl;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ElementValuePair;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

public class Factory {
  public static final Byte DUMMY_BYTE = Byte.valueOf((byte)0);
  
  public static final Character DUMMY_CHAR = Character.valueOf('0');
  
  public static final Double DUMMY_DOUBLE = Double.valueOf(0.0D);
  
  public static final Float DUMMY_FLOAT = Float.valueOf(0.0F);
  
  public static final Integer DUMMY_INTEGER = Integer.valueOf(0);
  
  public static final Long DUMMY_LONG = Long.valueOf(0L);
  
  public static final Short DUMMY_SHORT = Short.valueOf((short)0);
  
  private final BaseProcessingEnvImpl _env;
  
  public static List<? extends AnnotationMirror> EMPTY_ANNOTATION_MIRRORS = Collections.emptyList();
  
  public Factory(BaseProcessingEnvImpl env) {
    this._env = env;
  }
  
  public List<? extends AnnotationMirror> getAnnotationMirrors(AnnotationBinding[] annotations) {
    if (annotations == null || annotations.length == 0)
      return Collections.emptyList(); 
    List<AnnotationMirror> list = new ArrayList<>(annotations.length);
    byte b;
    int i;
    AnnotationBinding[] arrayOfAnnotationBinding;
    for (i = (arrayOfAnnotationBinding = annotations).length, b = 0; b < i; ) {
      AnnotationBinding annotation = arrayOfAnnotationBinding[b];
      if (annotation != null)
        list.add(newAnnotationMirror(annotation)); 
      b++;
    } 
    return Collections.unmodifiableList(list);
  }
  
  public <A extends Annotation> A[] getAnnotationsByType(AnnotationBinding[] annoInstances, Class<A> annotationClass) {
    Annotation[] result = getAnnotations(annoInstances, annotationClass, false);
    return (result == null) ? (A[])Array.newInstance(annotationClass, 0) : (A[])result;
  }
  
  public <A extends Annotation> A getAnnotation(AnnotationBinding[] annoInstances, Class<A> annotationClass) {
    Annotation[] result = getAnnotations(annoInstances, annotationClass, true);
    return (result == null) ? null : (A)result[0];
  }
  
  private <A extends Annotation> A[] getAnnotations(AnnotationBinding[] annoInstances, Class<A> annotationClass, boolean justTheFirst) {
    if (annoInstances == null || annoInstances.length == 0 || annotationClass == null)
      return null; 
    String annoTypeName = annotationClass.getName();
    if (annoTypeName == null)
      return null; 
    List<A> list = new ArrayList<>(annoInstances.length);
    byte b;
    int i;
    AnnotationBinding[] arrayOfAnnotationBinding;
    for (i = (arrayOfAnnotationBinding = annoInstances).length, b = 0; b < i; ) {
      AnnotationBinding annoInstance = arrayOfAnnotationBinding[b];
      if (annoInstance != null) {
        AnnotationMirrorImpl annoMirror = createAnnotationMirror(annoTypeName, annoInstance);
        if (annoMirror != null) {
          list.add((A)Proxy.newProxyInstance(annotationClass.getClassLoader(), new Class[] { annotationClass }, (InvocationHandler)annoMirror));
          if (justTheFirst)
            break; 
        } 
      } 
      b++;
    } 
    Annotation[] result = (Annotation[])Array.newInstance(annotationClass, list.size());
    return (list.size() > 0) ? (A[])list.<Annotation>toArray(result) : null;
  }
  
  private AnnotationMirrorImpl createAnnotationMirror(String annoTypeName, AnnotationBinding annoInstance) {
    ReferenceBinding binding = annoInstance.getAnnotationType();
    if (binding != null && binding.isAnnotationType()) {
      char[] qName;
      if (binding.isMemberType()) {
        annoTypeName = annoTypeName.replace('$', '.');
        qName = CharOperation.concatWith((binding.enclosingType()).compoundName, binding.sourceName, '.');
        CharOperation.replace(qName, '$', '.');
      } else {
        qName = CharOperation.concatWith(binding.compoundName, '.');
      } 
      if (annoTypeName.equals(new String(qName)))
        return (AnnotationMirrorImpl)this._env.getFactory().newAnnotationMirror(annoInstance); 
    } 
    return null;
  }
  
  private static void appendModifier(Set<Modifier> result, int modifiers, int modifierConstant, Modifier modifier) {
    if ((modifiers & modifierConstant) != 0)
      result.add(modifier); 
  }
  
  private static void decodeModifiers(Set<Modifier> result, int modifiers, int[] checkBits) {
    if (checkBits == null)
      return; 
    for (int i = 0, max = checkBits.length; i < max; i++) {
      switch (checkBits[i]) {
        case 1:
          appendModifier(result, modifiers, checkBits[i], Modifier.PUBLIC);
          break;
        case 4:
          appendModifier(result, modifiers, checkBits[i], Modifier.PROTECTED);
          break;
        case 2:
          appendModifier(result, modifiers, checkBits[i], Modifier.PRIVATE);
          break;
        case 1024:
          appendModifier(result, modifiers, checkBits[i], Modifier.ABSTRACT);
          break;
        case 65536:
          try {
            appendModifier(result, modifiers, checkBits[i], Modifier.valueOf("DEFAULT"));
          } catch (IllegalArgumentException illegalArgumentException) {}
          break;
        case 8:
          appendModifier(result, modifiers, checkBits[i], Modifier.STATIC);
          break;
        case 16:
          appendModifier(result, modifiers, checkBits[i], Modifier.FINAL);
          break;
        case 32:
          appendModifier(result, modifiers, checkBits[i], Modifier.SYNCHRONIZED);
          break;
        case 256:
          appendModifier(result, modifiers, checkBits[i], Modifier.NATIVE);
          break;
        case 2048:
          appendModifier(result, modifiers, checkBits[i], Modifier.STRICTFP);
          break;
        case 128:
          appendModifier(result, modifiers, checkBits[i], Modifier.TRANSIENT);
          break;
        case 64:
          appendModifier(result, modifiers, checkBits[i], Modifier.VOLATILE);
          break;
      } 
    } 
  }
  
  public static Object getMatchingDummyValue(Class<?> expectedType) {
    if (expectedType.isPrimitive()) {
      if (expectedType == boolean.class)
        return Boolean.FALSE; 
      if (expectedType == byte.class)
        return DUMMY_BYTE; 
      if (expectedType == char.class)
        return DUMMY_CHAR; 
      if (expectedType == double.class)
        return DUMMY_DOUBLE; 
      if (expectedType == float.class)
        return DUMMY_FLOAT; 
      if (expectedType == int.class)
        return DUMMY_INTEGER; 
      if (expectedType == long.class)
        return DUMMY_LONG; 
      if (expectedType == short.class)
        return DUMMY_SHORT; 
      return DUMMY_INTEGER;
    } 
    return null;
  }
  
  public TypeMirror getReceiverType(MethodBinding binding) {
    if (binding != null) {
      if (binding.receiver != null)
        return this._env.getFactory().newTypeMirror((Binding)binding.receiver); 
      if (binding.declaringClass != null && 
        !binding.isStatic() && (!binding.isConstructor() || binding.declaringClass.isMemberType()))
        return this._env.getFactory().newTypeMirror((Binding)binding.declaringClass); 
    } 
    return NoTypeImpl.NO_TYPE_NONE;
  }
  
  public static Set<Modifier> getModifiers(int modifiers, ElementKind kind) {
    return getModifiers(modifiers, kind, false);
  }
  
  public static Set<Modifier> getModifiers(int modifiers, ElementKind kind, boolean isFromBinary) {
    EnumSet<Modifier> result = EnumSet.noneOf(Modifier.class);
    switch (kind) {
      case METHOD:
      case CONSTRUCTOR:
        decodeModifiers(result, modifiers, new int[] { 1, 
              4, 
              2, 
              1024, 
              8, 
              16, 
              32, 
              256, 
              2048, 
              65536 });
        break;
      case ENUM_CONSTANT:
      case FIELD:
        decodeModifiers(result, modifiers, new int[] { 1, 
              4, 
              2, 
              8, 
              16, 
              128, 
              64 });
        break;
      case ENUM:
        if (isFromBinary) {
          decodeModifiers(result, modifiers, new int[] { 1, 
                4, 
                16, 
                2, 
                1024, 
                8, 
                2048 });
          break;
        } 
        decodeModifiers(result, modifiers, new int[] { 1, 
              4, 
              16, 
              2, 
              8, 
              2048 });
        break;
      case CLASS:
      case null:
      case INTERFACE:
        decodeModifiers(result, modifiers, new int[] { 1, 
              4, 
              1024, 
              16, 
              2, 
              8, 
              2048 });
        break;
    } 
    return Collections.unmodifiableSet(result);
  }
  
  public AnnotationMirror newAnnotationMirror(AnnotationBinding binding) {
    return (AnnotationMirror)new AnnotationMirrorImpl(this._env, binding);
  }
  
  public Element newElement(Binding binding, ElementKind kindHint) {
    ReferenceBinding referenceBinding;
    if (binding == null)
      return null; 
    switch (binding.kind()) {
      case 1:
      case 2:
      case 3:
        return (Element)new VariableElementImpl(this._env, (VariableBinding)binding);
      case 4:
      case 2052:
        referenceBinding = (ReferenceBinding)binding;
        if ((referenceBinding.tagBits & 0x80L) != 0L)
          return (Element)new ErrorTypeElement(this._env, referenceBinding); 
        if (CharOperation.equals(referenceBinding.sourceName, TypeConstants.PACKAGE_INFO_NAME))
          return (Element)new PackageElementImpl(this._env, referenceBinding.fPackage); 
        return (Element)new TypeElementImpl(this._env, referenceBinding, kindHint);
      case 8:
        return (Element)new ExecutableElementImpl(this._env, (MethodBinding)binding);
      case 260:
      case 1028:
        return (Element)new TypeElementImpl(this._env, ((ParameterizedTypeBinding)binding).genericType(), kindHint);
      case 16:
        return (Element)new PackageElementImpl(this._env, (PackageBinding)binding);
      case 4100:
        return (Element)new TypeParameterElementImpl(this._env, (TypeVariableBinding)binding);
      case 32:
      case 68:
      case 132:
      case 516:
      case 8196:
        throw new UnsupportedOperationException("NYI: binding type " + binding.kind());
    } 
    return null;
  }
  
  public Element newElement(Binding binding) {
    return newElement(binding, null);
  }
  
  public PackageElement newPackageElement(PackageBinding binding) {
    return (PackageElement)new PackageElementImpl(this._env, binding);
  }
  
  public NullType getNullType() {
    return NoTypeImpl.NULL_TYPE;
  }
  
  public NoType getNoType(TypeKind kind) {
    switch (kind) {
      case NONE:
        return NoTypeImpl.NO_TYPE_NONE;
      case VOID:
        return NoTypeImpl.NO_TYPE_VOID;
      case PACKAGE:
        return NoTypeImpl.NO_TYPE_PACKAGE;
    } 
    throw new IllegalArgumentException();
  }
  
  public PrimitiveTypeImpl getPrimitiveType(TypeKind kind) {
    switch (kind) {
      case BOOLEAN:
        return PrimitiveTypeImpl.BOOLEAN;
      case BYTE:
        return PrimitiveTypeImpl.BYTE;
      case CHAR:
        return PrimitiveTypeImpl.CHAR;
      case DOUBLE:
        return PrimitiveTypeImpl.DOUBLE;
      case FLOAT:
        return PrimitiveTypeImpl.FLOAT;
      case INT:
        return PrimitiveTypeImpl.INT;
      case LONG:
        return PrimitiveTypeImpl.LONG;
      case SHORT:
        return PrimitiveTypeImpl.SHORT;
    } 
    throw new IllegalArgumentException();
  }
  
  public PrimitiveTypeImpl getPrimitiveType(BaseTypeBinding binding) {
    AnnotationBinding[] annotations = binding.getTypeAnnotations();
    if (annotations == null || annotations.length == 0)
      return getPrimitiveType(PrimitiveTypeImpl.getKind(binding)); 
    return new PrimitiveTypeImpl(this._env, binding);
  }
  
  public TypeMirror newTypeMirror(Binding binding) {
    ReferenceBinding referenceBinding;
    BaseTypeBinding btb;
    switch (binding.kind()) {
      case 1:
      case 2:
      case 3:
        return newTypeMirror((Binding)((VariableBinding)binding).type);
      case 16:
        return getNoType(TypeKind.PACKAGE);
      case 32:
        throw new UnsupportedOperationException("NYI: import type " + binding.kind());
      case 8:
        return (TypeMirror)new ExecutableTypeImpl(this._env, (MethodBinding)binding);
      case 4:
      case 260:
      case 1028:
      case 2052:
        referenceBinding = (ReferenceBinding)binding;
        if ((referenceBinding.tagBits & 0x80L) != 0L)
          return getErrorType(referenceBinding); 
        return (TypeMirror)new DeclaredTypeImpl(this._env, (ReferenceBinding)binding);
      case 68:
        return (TypeMirror)new ArrayTypeImpl(this._env, (ArrayBinding)binding);
      case 132:
        btb = (BaseTypeBinding)binding;
        switch (btb.id) {
          case 6:
            return getNoType(TypeKind.VOID);
          case 12:
            return getNullType();
        } 
        return (TypeMirror)getPrimitiveType(btb);
      case 516:
      case 8196:
        return (TypeMirror)new WildcardTypeImpl(this._env, (WildcardBinding)binding);
      case 4100:
        return (TypeMirror)new TypeVariableImpl(this._env, (TypeVariableBinding)binding);
    } 
    return null;
  }
  
  public TypeParameterElement newTypeParameterElement(TypeVariableBinding variable, Element declaringElement) {
    return (TypeParameterElement)new TypeParameterElementImpl(this._env, variable, declaringElement);
  }
  
  public ErrorType getErrorType(ReferenceBinding binding) {
    return (ErrorType)new ErrorTypeImpl(this._env, binding);
  }
  
  public static Object performNecessaryPrimitiveTypeConversion(Class<?> expectedType, Object value, boolean avoidReflectException) {
    assert expectedType.isPrimitive() : "expectedType is not a primitive type: " + expectedType.getName();
    if (value == null)
      return avoidReflectException ? getMatchingDummyValue(expectedType) : null; 
    String typeName = expectedType.getName();
    char expectedTypeChar = typeName.charAt(0);
    int nameLen = typeName.length();
    if (value instanceof Byte) {
      byte b = ((Byte)value).byteValue();
      switch (expectedTypeChar) {
        case 'b':
          if (nameLen == 4)
            return value; 
          return avoidReflectException ? Boolean.FALSE : value;
        case 'c':
          return Character.valueOf((char)b);
        case 'd':
          return new Double(b);
        case 'f':
          return new Float(b);
        case 'i':
          return Integer.valueOf(b);
        case 'l':
          return Long.valueOf(b);
        case 's':
          return Short.valueOf(b);
      } 
      throw new IllegalStateException("unknown type " + expectedTypeChar);
    } 
    if (value instanceof Short) {
      short s = ((Short)value).shortValue();
      switch (expectedTypeChar) {
        case 'b':
          if (nameLen == 4)
            return Byte.valueOf((byte)s); 
          return avoidReflectException ? Boolean.FALSE : value;
        case 'c':
          return Character.valueOf((char)s);
        case 'd':
          return new Double(s);
        case 'f':
          return new Float(s);
        case 'i':
          return Integer.valueOf(s);
        case 'l':
          return Long.valueOf(s);
        case 's':
          return value;
      } 
      throw new IllegalStateException("unknown type " + expectedTypeChar);
    } 
    if (value instanceof Character) {
      char c = ((Character)value).charValue();
      switch (expectedTypeChar) {
        case 'b':
          if (nameLen == 4)
            return Byte.valueOf((byte)c); 
          return avoidReflectException ? Boolean.FALSE : value;
        case 'c':
          return value;
        case 'd':
          return new Double(c);
        case 'f':
          return new Float(c);
        case 'i':
          return Integer.valueOf(c);
        case 'l':
          return Long.valueOf(c);
        case 's':
          return Short.valueOf((short)c);
      } 
      throw new IllegalStateException("unknown type " + expectedTypeChar);
    } 
    if (value instanceof Integer) {
      int i = ((Integer)value).intValue();
      switch (expectedTypeChar) {
        case 'b':
          if (nameLen == 4)
            return Byte.valueOf((byte)i); 
          return avoidReflectException ? Boolean.FALSE : value;
        case 'c':
          return Character.valueOf((char)i);
        case 'd':
          return new Double(i);
        case 'f':
          return new Float(i);
        case 'i':
          return value;
        case 'l':
          return Long.valueOf(i);
        case 's':
          return Short.valueOf((short)i);
      } 
      throw new IllegalStateException("unknown type " + expectedTypeChar);
    } 
    if (value instanceof Long) {
      long l = ((Long)value).longValue();
      switch (expectedTypeChar) {
        case 'b':
        case 'c':
        case 'i':
        case 's':
          return avoidReflectException ? getMatchingDummyValue(expectedType) : value;
        case 'd':
          return new Double(l);
        case 'f':
          return new Float((float)l);
        case 'l':
          return value;
      } 
      throw new IllegalStateException("unknown type " + expectedTypeChar);
    } 
    if (value instanceof Float) {
      float f = ((Float)value).floatValue();
      switch (expectedTypeChar) {
        case 'b':
        case 'c':
        case 'i':
        case 'l':
        case 's':
          return avoidReflectException ? getMatchingDummyValue(expectedType) : value;
        case 'd':
          return new Double(f);
        case 'f':
          return value;
      } 
      throw new IllegalStateException("unknown type " + expectedTypeChar);
    } 
    if (value instanceof Double) {
      if (expectedTypeChar == 'd')
        return value; 
      return avoidReflectException ? getMatchingDummyValue(expectedType) : value;
    } 
    if (value instanceof Boolean) {
      if (expectedTypeChar == 'b' && nameLen == 7)
        return value; 
      return avoidReflectException ? getMatchingDummyValue(expectedType) : value;
    } 
    return avoidReflectException ? getMatchingDummyValue(expectedType) : value;
  }
  
  public static void setArrayMatchingDummyValue(Object array, int i, Class<?> expectedLeafType) {
    if (boolean.class.equals(expectedLeafType)) {
      Array.setBoolean(array, i, false);
    } else if (byte.class.equals(expectedLeafType)) {
      Array.setByte(array, i, DUMMY_BYTE.byteValue());
    } else if (char.class.equals(expectedLeafType)) {
      Array.setChar(array, i, DUMMY_CHAR.charValue());
    } else if (double.class.equals(expectedLeafType)) {
      Array.setDouble(array, i, DUMMY_DOUBLE.doubleValue());
    } else if (float.class.equals(expectedLeafType)) {
      Array.setFloat(array, i, DUMMY_FLOAT.floatValue());
    } else if (int.class.equals(expectedLeafType)) {
      Array.setInt(array, i, DUMMY_INTEGER.intValue());
    } else if (long.class.equals(expectedLeafType)) {
      Array.setLong(array, i, DUMMY_LONG.longValue());
    } else if (short.class.equals(expectedLeafType)) {
      Array.setShort(array, i, DUMMY_SHORT.shortValue());
    } else {
      Array.set(array, i, null);
    } 
  }
  
  public static AnnotationBinding[] getPackedAnnotationBindings(AnnotationBinding[] annotations) {
    int length = (annotations == null) ? 0 : annotations.length;
    if (length == 0)
      return annotations; 
    AnnotationBinding[] repackagedBindings = annotations;
    for (int i = 0; i < length; i++) {
      AnnotationBinding annotation = repackagedBindings[i];
      if (annotation != null) {
        ReferenceBinding annotationType = annotation.getAnnotationType();
        if (annotationType.isRepeatableAnnotationType()) {
          ReferenceBinding containerType = annotationType.containerAnnotationType();
          if (containerType != null) {
            MethodBinding[] values = containerType.getMethods(TypeConstants.VALUE);
            if (values != null && values.length == 1) {
              MethodBinding value = values[0];
              if (value.returnType != null && value.returnType.dimensions() == 1 && !TypeBinding.notEquals(value.returnType.leafComponentType(), (TypeBinding)annotationType)) {
                List<AnnotationBinding> containees = null;
                for (int m = i + 1; m < length; m++) {
                  AnnotationBinding otherAnnotation = repackagedBindings[m];
                  if (otherAnnotation != null && 
                    otherAnnotation.getAnnotationType() == annotationType) {
                    if (repackagedBindings == annotations)
                      System.arraycopy(repackagedBindings, 0, repackagedBindings = new AnnotationBinding[length], 0, length); 
                    repackagedBindings[m] = null;
                    if (containees == null) {
                      containees = new ArrayList<>();
                      containees.add(annotation);
                    } 
                    containees.add(otherAnnotation);
                  } 
                } 
                if (containees != null) {
                  ElementValuePair[] elementValuePairs = { new ElementValuePair(TypeConstants.VALUE, containees.toArray(), value) };
                  repackagedBindings[i] = new AnnotationBinding(containerType, elementValuePairs);
                } 
              } 
            } 
          } 
        } 
      } 
    } 
    int finalTally = 0;
    int k;
    for (k = 0; k < length; k++) {
      if (repackagedBindings[k] != null)
        finalTally++; 
    } 
    if (repackagedBindings == annotations && finalTally == length)
      return annotations; 
    annotations = new AnnotationBinding[finalTally];
    int j;
    for (k = 0, j = 0; k < length; k++) {
      if (repackagedBindings[k] != null)
        annotations[j++] = repackagedBindings[k]; 
    } 
    return annotations;
  }
  
  public static AnnotationBinding[] getUnpackedAnnotationBindings(AnnotationBinding[] annotations) {
    int length = (annotations == null) ? 0 : annotations.length;
    if (length == 0)
      return annotations; 
    List<AnnotationBinding> unpackedAnnotations = new ArrayList<>();
    for (int i = 0; i < length; i++) {
      AnnotationBinding annotation = annotations[i];
      if (annotation != null) {
        unpackedAnnotations.add(annotation);
        ReferenceBinding annotationType = annotation.getAnnotationType();
        MethodBinding[] values = annotationType.getMethods(TypeConstants.VALUE);
        if (values != null && values.length == 1) {
          MethodBinding value = values[0];
          if (value.returnType.dimensions() == 1) {
            TypeBinding containeeType = value.returnType.leafComponentType();
            if (containeeType != null && containeeType.isAnnotationType() && containeeType.isRepeatableAnnotationType())
              if (containeeType.containerAnnotationType() == annotationType) {
                ElementValuePair[] elementValuePairs = annotation.getElementValuePairs();
                byte b;
                int j;
                ElementValuePair[] arrayOfElementValuePair1;
                for (j = (arrayOfElementValuePair1 = elementValuePairs).length, b = 0; b < j; ) {
                  ElementValuePair elementValuePair = arrayOfElementValuePair1[b];
                  if (CharOperation.equals(elementValuePair.getName(), TypeConstants.VALUE)) {
                    Object[] containees = (Object[])elementValuePair.getValue();
                    byte b1;
                    int k;
                    Object[] arrayOfObject1;
                    for (k = (arrayOfObject1 = containees).length, b1 = 0; b1 < k; ) {
                      Object object = arrayOfObject1[b1];
                      unpackedAnnotations.add((AnnotationBinding)object);
                      b1++;
                    } 
                    break;
                  } 
                  b++;
                } 
              }  
          } 
        } 
      } 
    } 
    return unpackedAnnotations.<AnnotationBinding>toArray(new AnnotationBinding[unpackedAnnotations.size()]);
  }
}
